/*
 * ViPer Applets Series
 * binshape.java -- binocular shape reconstruction
 * Copyright (C) 2000 Filip Pizlo
 *
 * Uses ccWizard, (C) 2000 Lyceum Technologies
 *
 * Data arrays:
 * -> Points for the original cube
 * -> Points for reconstructed object
 * -> Eye-point vector (two of them)
 * -> Vector orthogonal to retina, pointing at cetner of object (two of them)
 * -> Points for left and right images
 * -> Vector to rotate retinal plane (two)
 */

import java.awt.*;
import java.awt.event.*;
import java.util.*;
import java.applet.*;

public class binshape extends ccWizard {
 int state;
 Random rand=new Random();
 double
  interOcularDist=7.0,
  cubeSize=10.0,
  rightEye=1.0,
  leftEye=-1.0,
  xTrans,
  yTrans,
  zTrans,
  stdDev,
  elevationError,
  vergenceError,
  torsionError;
 double[]
  xAxis={1.0,0.0,0.0},
  yAxis={0.0,1.0,0.0},
  zAxis={0.0,0.0,1.0};
 CheckboxGroup selection;
 Checkbox noError,stdError,custError;
 TextField xTransField,yTransField,zTransField,sizeField,stdDevField,
  elevationField,vergenceField,torsionField;
 Label xTransLabel,yTransLabel,zTransLabel,sizeLabel,stdDevLabel,
  elevationLabel,vergenceLabel,torsionLabel;
 int[] coolStuff={100,200,300,400,500};
 public void setDefaultValues() {
  xTrans=0.0;
  yTrans=0.0;
  zTrans=cmToUnit(coolStuff[Math.abs(rand.nextInt())%coolStuff.length]);
  cubeSize=2.0;
  stdDev=0.02;
  elevationError=0.0;
  vergenceError=0.0;
  torsionError=0.0;
 }
 public void setPerfectValues() {
  stdDev=0;
  elevationError=0.0;
  vergenceError=0.0;
  torsionError=0.0;
 }
 public void updateValuesToFields() {
  xTransField.setText(""+unitToCm(xTrans));
  yTransField.setText(""+unitToCm(yTrans));
  zTransField.setText(""+unitToCm(zTrans));
  sizeField.setText(""+unitToCm(cubeSize));
  stdDevField.setText(""+(stdDev*100.0));
  elevationField.setText(""+(elevationError*180.0/Math.PI));
  vergenceField.setText(""+(vergenceError*180.0/Math.PI));
  torsionField.setText(""+(torsionError*180.0/Math.PI));
 }
 public void updateFieldsToValues() {
  xTrans=cmToUnit(Double.valueOf(xTransField.getText()).doubleValue());
  yTrans=cmToUnit(Double.valueOf(yTransField.getText()).doubleValue());
  zTrans=cmToUnit(Double.valueOf(zTransField.getText()).doubleValue());
  cubeSize=cmToUnit(Double.valueOf(sizeField.getText()).doubleValue());
  stdDev=Double.valueOf(stdDevField.getText()).doubleValue()/100.0;
  elevationError=
   Double.valueOf(elevationField.getText()).doubleValue()*Math.PI/180.0;
  vergenceError=
   Double.valueOf(vergenceField.getText()).doubleValue()*Math.PI/180.0;
  torsionError=
   Double.valueOf(torsionField.getText()).doubleValue()*Math.PI/180.0;
 }
 public void updateFieldsToLabels() {
  xTransLabel.setText("X-translation = "+xTransField.getText()+" cm");
  yTransLabel.setText("Y-translation = "+yTransField.getText()+" cm");
  zTransLabel.setText("Z-translation = "+zTransField.getText()+" cm");
  sizeLabel.setText("Cube Size = "+sizeField.getText()+" cm");
  stdDevLabel.setText("Gaussian Noise = "+stdDevField.getText()+" %");
  elevationLabel.setText("Elevation Error = "+elevationField.getText()+
   " degrees");
  vergenceLabel.setText("Vergence Error = "+vergenceField.getText()+
   " degrees");
  torsionLabel.setText("Torsion Error = "+torsionField.getText()+" degrees");
 }
 public void init() {
  selection=new CheckboxGroup();
  noError=new Checkbox("Base Case - No Error",selection,false);
  stdError=new Checkbox("Default Case",selection,true);
  custError=new Checkbox("User Specified",selection,false);
  xTransField=new TextField(20);
  yTransField=new TextField(20);
  zTransField=new TextField(20);
  sizeField=new TextField(20);
  stdDevField=new TextField(20);
  elevationField=new TextField(20);
  vergenceField=new TextField(20);
  torsionField=new TextField(20);
  xTransLabel=new Label();
  yTransLabel=new Label();
  zTransLabel=new Label();
  sizeLabel=new Label();
  stdDevLabel=new Label();
  elevationLabel=new Label();
  vergenceLabel=new Label();
  torsionLabel=new Label();
  setBackground(new Color(206,206,206));
  super.init();
  done.setText("Reset");
  Panel inPanE=new Panel();
  Panel inPanE_text=new Panel(new GridLayout(0,1));
  inPanE_text.add(new Label("The applet has experienced an error."));
  inPanE_text.add(new Label(""));
  inPanE_text.add(new Label("Please report this error to"));
  inPanE_text.add(new Label("pizlo@purdue.edu along with all the Java"));
  inPanE_text.add(new Label("log, browser version, and OS version."));
  inPanE_text.add(new Label(""));
  inPanE_text.add(new Label("To continue using the applet, click the"));
  inPanE_text.add(new Label("\"Reload\" button on your browser."));
  inPanE.add(inPanE_text);
  centPan.add("error",inPanE);
  Panel inPan0=new Panel();
  Panel inPan0_text=new Panel(new GridLayout(0,1));
  inPan0_text.add(new Label("Welcome to the Binocular Shape Reconstruction"));
  inPan0_text.add(new Label("Java-Applet!"));
  inPan0_text.add(new Label(""));
  inPan0_text.add(new Label("This applet demonstrates the effect of visual"));
  inPan0_text.add(new Label("and oculomotor noise (error) on 3D shape"));
  inPan0_text.add(new Label("reconstruction, when this reconstruction is"));
  inPan0_text.add(new Label("based solely on visual and oculomotor"));
  inPan0_text.add(new Label("information."));
  inPan0_text.add(new Label(""));
  inPan0_text.add(new Label("Click \"Next\" to get started!"));
  inPan0.add(inPan0_text);
  centPan.add("first",inPan0);
  Panel inPan1=new Panel();
  Panel inPan1_pre=new Panel(new BorderLayout(3,3));
  Panel inPan1_text=new Panel(new GridLayout(0,1));
  inPan1_text.add(new Label("Select the parameters of the reconstruction."));
  inPan1_text.add(new Label(""));
  Panel inPan1_selPan=new Panel(new GridLayout(0,1));
  Panel inPan1_noPan=new Panel(new BorderLayout());
  inPan1_noPan.add("West",noError);
  inPan1_noPan.add("Center",new Panel());
  Panel inPan1_stdPan=new Panel(new BorderLayout());
  inPan1_stdPan.add("West",stdError);
  inPan1_stdPan.add("Center",new Panel());
  Panel inPan1_custPan=new Panel(new BorderLayout());
  inPan1_custPan.add("West",custError);
  inPan1_custPan.add("Center",new Panel());
  //inPan1_selPan.add(inPan1_noPan);
  inPan1_selPan.add(inPan1_stdPan);
  inPan1_selPan.add(inPan1_custPan);
  Panel inPan1_selPanPre=new Panel(new BorderLayout());
  inPan1_selPanPre.add("North",inPan1_selPan);
  inPan1_selPanPre.add("Center",new Panel());
  inPan1_pre.add("North",inPan1_text);
  inPan1_pre.add("Center",inPan1_selPanPre);
  inPan1.add(inPan1_pre);
  centPan.add("second",inPan1);
  Panel inPanEnt1=new Panel();
  Panel inPanEnt1_pre=new Panel(new BorderLayout(5,5));
  Panel inPanEnt1_text=new Panel(new GridLayout(0,1));
  inPanEnt1_text.add(new Label("Step 1: Translation Values"));
  inPanEnt1_text.add(
   new Label("Position of the cube relative to the observer:"));
  inPanEnt1_text.add(new Label("X - horizontal, Y - vertical, Z - depth."));
  inPanEnt1_text.add(new Label(""));
  Panel inPanEnt1_mainPre=new Panel();
  Panel inPanEnt1_main=new Panel(new BorderLayout(3,3));
  Panel inPanEnt1_farRight=new Panel(new GridLayout(0,1,3,3));
  inPanEnt1_farRight.add(new Label("cm"));
  inPanEnt1_farRight.add(new Label("cm"));
  inPanEnt1_farRight.add(new Label("cm"));
  Panel inPanEnt1_right=new Panel(new GridLayout(0,1,3,3));
  inPanEnt1_right.add(xTransField);
  inPanEnt1_right.add(yTransField);
  inPanEnt1_right.add(zTransField);
  Panel inPanEnt1_left=new Panel(new GridLayout(0,1,3,3));
  inPanEnt1_left.add(new Label("X-translation:"));
  inPanEnt1_left.add(new Label("Y-translation:"));
  inPanEnt1_left.add(new Label("Z-translation:"));
  inPanEnt1_main.add("West",inPanEnt1_left);
  inPanEnt1_main.add("Center",inPanEnt1_right);
  inPanEnt1_main.add("East",inPanEnt1_farRight);
  inPanEnt1_mainPre.add(inPanEnt1_main);
  inPanEnt1_pre.add("North",inPanEnt1_text);
  inPanEnt1_pre.add("Center",inPanEnt1_mainPre);
  inPanEnt1.add(inPanEnt1_pre);
  centPan.add("entFirst",inPanEnt1);
  Panel inPanEnt2=new Panel();
  Panel inPanEnt2_pre=new Panel(new BorderLayout(5,5));
  Panel inPanEnt2_text=new Panel(new GridLayout(0,1));
  inPanEnt2_text.add(new Label("Step 2: Cube Size"));
  inPanEnt2_text.add(
   new Label("Enter the size, in centimeters, that the cube"));
  inPanEnt2_text.add(new Label("should have."));
  inPanEnt2_text.add(new Label(""));
  Panel inPanEnt2_mainPre=new Panel();
  Panel inPanEnt2_main=new Panel(new BorderLayout(3,3));
  inPanEnt2_main.add("West",new Label("Cube Size:"));
  inPanEnt2_main.add("Center",sizeField);
  inPanEnt2_main.add("East",new Label("cm"));
  inPanEnt2_mainPre.add(inPanEnt2_main);
  inPanEnt2_pre.add("North",inPanEnt2_text);
  inPanEnt2_pre.add("Center",inPanEnt2_mainPre);
  inPanEnt2.add(inPanEnt2_pre);
  centPan.add("entSecond",inPanEnt2);
  Panel inPanEnt3=new Panel();
  Panel inPanEnt3_pre=new Panel(new BorderLayout(5,5));
  Panel inPanEnt3_text=new Panel(new GridLayout(0,1));
  inPanEnt3_text.add(new Label("Step 3: Visual Noise"));
  inPanEnt3_text.add(new Label("Please specify the standard deviation of"));
  inPanEnt3_text.add(new Label("Gaussian noise.  This noise will be imposed"));
  inPanEnt3_text.add(new Label("on the retinal images of the vertices of the"));
  inPanEnt3_text.add(new Label("cube."));
  inPanEnt3_text.add(new Label(""));
  Panel inPanEnt3_mainPre=new Panel();
  Panel inPanEnt3_main=new Panel(new BorderLayout(3,3));
  inPanEnt3_main.add("West",new Label("Gaussian Noise:"));
  inPanEnt3_main.add("Center",stdDevField);
  inPanEnt3_main.add("East",new Label("%"));
  inPanEnt3_mainPre.add(inPanEnt3_main);
  inPanEnt3_pre.add("North",inPanEnt3_text);
  inPanEnt3_pre.add("Center",inPanEnt3_mainPre);
  inPanEnt3.add(inPanEnt3_pre);
  centPan.add("entThird",inPanEnt3);
  Panel inPanEnt4=new Panel();
  Panel inPanEnt4_pre=new Panel(new BorderLayout(5,5));
  Panel inPanEnt4_text=new Panel(new GridLayout(0,1));
  inPanEnt4_text.add(new Label("Step 4: Oculomotor Errors"));
  inPanEnt4_text.add(new Label("Please enter the error values below."));
  inPanEnt4_text.add(new Label(""));
  Panel inPanEnt4_mainPre=new Panel();
  Panel inPanEnt4_main=new Panel(new BorderLayout(3,3));
  Panel inPanEnt4_farRight=new Panel(new GridLayout(0,1,3,3));
  inPanEnt4_farRight.add(new Label("degrees"));
  inPanEnt4_farRight.add(new Label("degrees"));
  inPanEnt4_farRight.add(new Label("degrees"));
  Panel inPanEnt4_right=new Panel(new GridLayout(0,1,3,3));
  inPanEnt4_right.add(elevationField);
  inPanEnt4_right.add(vergenceField);
  inPanEnt4_right.add(torsionField);
  Panel inPanEnt4_left=new Panel(new GridLayout(0,1,3,3));
  inPanEnt4_left.add(new Label("Elevation:"));
  inPanEnt4_left.add(new Label("Vergence:"));
  inPanEnt4_left.add(new Label("Torsion:"));
  inPanEnt4_main.add("West",inPanEnt4_left);
  inPanEnt4_main.add("Center",inPanEnt4_right);
  inPanEnt4_main.add("East",inPanEnt4_farRight);
  inPanEnt4_mainPre.add(inPanEnt4_main);
  inPanEnt4_pre.add("North",inPanEnt4_text);
  inPanEnt4_pre.add("Center",inPanEnt4_mainPre);
  inPanEnt4.add(inPanEnt4_pre);
  centPan.add("entFourth",inPanEnt4);
  Panel inPanPre=new Panel();
  Panel inPanPre_text=new Panel(new GridLayout(0,1));
  inPanPre_text.add(
   new Label("The animation of the skewed cube should appear"));
  inPanPre_text.add(new Label("in a separate window."));
  inPanPre_text.add(new Label(""));
  inPanPre_text.add(new Label("Below are the values:"));
  inPanPre_text.add(xTransLabel);
  inPanPre_text.add(yTransLabel);
  inPanPre_text.add(zTransLabel);
  inPanPre_text.add(sizeLabel);
  inPanPre_text.add(stdDevLabel);
  inPanPre_text.add(elevationLabel);
  inPanPre_text.add(vergenceLabel);
  inPanPre_text.add(torsionLabel);
  inPanPre.add(inPanPre_text);
  centPan.add("preRotation",inPanPre);
  Panel inPanRot=new Panel();
  Panel inPanRot_text=new Panel(new GridLayout(0,1));
  inPanRot_text.add(
   new Label("You will see an animation of the original cube"));
  inPanRot_text.add(
   new Label("in a separate window.  Use the scrollbar to control"));
  inPanRot_text.add(new Label("speed."));
  inPanRot_text.add(new Label(""));
  inPanRot_text.add(new Label("Click \"Next\" to continue..."));
  inPanRot.add(inPanRot_text);
  centPan.add("rotation",inPanRot);
 }
 public synchronized void start() {
  // reset all components
  setDefaultValues();
  updateValuesToFields();
  state=0;
  analyzeState();
 }
 public synchronized void stop() {
  // do nothing
 }
 double[][] origCube;
 public synchronized void analyzeState() {
  try {
   enableButton(done);
   switch (state) {
    case 0:
     disableButton(prev);
     enableButton(next);
     canv.setImage(getImage(getCodeBase(),"binshapelogo.gif"));
     canv.repaint();
     card.show(centPan,"first");
     break;
    case 1:
     enableButton(prev);
     enableButton(next);
     canv.setImage(getImage(getCodeBase(),"binshapelogo.gif"));
     canv.repaint();
     card.show(centPan,"second");
     break;
    case 20:
     enableButton(prev);
     enableButton(next);
     canv.setImage(getImage(getCodeBase(),"binshapelogo.gif"));
     canv.repaint();
     card.show(centPan,"rotation");
     break;
    case 19:
     enableButton(prev);
     disableButton(next);
     canv.setImage(getImage(getCodeBase(),"binshapelogo.gif"));
     canv.repaint();
     origCube=randomizeOrientation(createCube());
     logCube("origCube",origCube);
     //double[][] cube=translateBy(scaleBy(origCube,
     // cmToUnit(cubeSize)),xTrans,yTrans,zTrans);
     System.out.println("xTrans="+xTrans+" yTrans="+yTrans+" zTrans="+zTrans);
     double[][] cube=translateBy(scaleBy(origCube,cubeSize),
      xTrans,yTrans,zTrans);
     logCube("cube",cube);
     double[] rightAxis=createVisualAxis(rightEye,xTrans,yTrans,zTrans);
     double[] leftAxis=createVisualAxis(leftEye,xTrans,yTrans,zTrans);
     double rightPlane=computeRetinalPlane(rightAxis,rightEye);
     double leftPlane=computeRetinalPlane(leftAxis,leftEye);
     double[][] rightInter=findIntersects(rightAxis,rightPlane,rightEye,cube);
     double[][] leftInter=findIntersects(leftAxis,leftPlane,leftEye,cube);
     logCube("rightInter",rightInter);
     logCube("leftInter",leftInter);
     System.out.println("rightAxis=("+rightAxis[0]+","+rightAxis[1]+","+
      rightAxis[2]+")");
     System.out.println("leftAxis=("+leftAxis[0]+","+leftAxis[1]+","+
      leftAxis[2]+")");
     double[] rightRotAxis=computeRetinaRotationAxis(rightAxis,rightPlane);
     double[] leftRotAxis=computeRetinaRotationAxis(leftAxis,leftPlane);
     double rightRotAng=getRotationAngle(rightAxis);
     double leftRotAng=getRotationAngle(leftAxis);
     double[][] rightNewInter=translateAndRotate(rightInter,rightAxis,
      rightPlane,rightEye,rightRotAng,rightRotAxis);
     double[][] leftNewInter=translateAndRotate(leftInter,leftAxis,leftPlane,
      leftEye,leftRotAng,leftRotAxis);
     logCube("rightNewInter",rightNewInter);
     logCube("leftNewInter",leftNewInter);
     System.out.println("stdDev="+stdDev);
     System.out.println("elevationError="+elevationError);
     System.out.println("vergenceError="+vergenceError);
     System.out.println("torsionError="+torsionError);
     double[][] rightNoise=addNoise(rightNewInter,stdDev);
     logCube("rightNoise",rightNoise);
     double[][] leftNoise=addNoise(leftNewInter,stdDev);
     logCube("leftNoise",leftNoise);
     double[][] leftSuperNoise=
      addUniversalError(
       addUniversalError(
        addUniversalError(leftNoise,
        xAxis,elevationError),
       yAxis,vergenceError),
      zAxis,torsionError);
     logCube("leftSuperNoise",leftSuperNoise);
     twoimages two=new twoimages();
     two.setCube1(leftNoise);
     two.setCube2(rightNoise);
     two.show();
     double[][] rightBadInter=undoTranslateAndRotate(rightNoise,rightAxis,
      rightPlane,rightEye,rightRotAng,rightRotAxis);
     logCube("rightBadInter",rightBadInter);
     double[][] leftBadInter=undoTranslateAndRotate(leftSuperNoise,leftAxis,
      leftPlane,leftEye,leftRotAng,leftRotAxis);
     logCube("leftBadInter",leftBadInter);
     double[][] newCube=reconstructCube(leftBadInter,rightBadInter);
     logCube("newCube",newCube);
     rotation rot2=new rotation(this);
     rot2.setCube(newCube);
     card.show(centPan,"preRotation");
     rot2.show();
     rot2.start(true);
     break;
    case 100:
     enableButton(prev);
     enableButton(next);
     canv.setImage(getImage(getCodeBase(),"binshapelogo.gif"));
     canv.repaint();
     updateFieldsToLabels();
     card.show(centPan,"entFirst");
     break;
    case 101:
     enableButton(prev);
     enableButton(next);
     canv.setImage(getImage(getCodeBase(),"binshapelogo.gif"));
     canv.repaint();
     updateFieldsToLabels();
     card.show(centPan,"entSecond");
     break;
    case 102:
     enableButton(prev);
     enableButton(next);
     canv.setImage(getImage(getCodeBase(),"binshapelogo.gif"));
     canv.repaint();
     updateFieldsToLabels();
     card.show(centPan,"entThird");
     break;
    case 103:
     enableButton(prev);
     enableButton(next);
     canv.setImage(getImage(getCodeBase(),"binshapelogo.gif"));
     canv.repaint();
     updateFieldsToLabels();
     card.show(centPan,"entFourth");
     break;
    default:
     throw new Exception("Unknown state!");
   }
  } catch (Throwable e) {
   System.err.println("Error in analyzeState(): "+e);
   disableButton(prev);
   disableButton(next);
   disableButton(done);
   canv.setImage(getImage(getCodeBase(),"error.gif"));
   canv.repaint();
   card.show(centPan,"error");
   stop();
  }
 }
 public synchronized void nextClicked() {
  switch (state) {
   case 0:
    state=20;
    break;
   case 20:
    state=1;
    break;
   case 1:
    Checkbox it=selection.getSelectedCheckbox();
    if (it==noError) {
     setPerfectValues();
     updateValuesToFields();
     updateFieldsToLabels();
     state=19;
    } else if (it==stdError) {
     updateValuesToFields();
     updateFieldsToLabels();
     state=19;
    } else if (it==custError) {
     state=100;
    }
    break;
   case 103:
    updateFieldsToValues();
    updateFieldsToLabels();
    state=19;
    break;
   default:
    state++;
    break;
  }
  analyzeState();
 }
 public synchronized void prevClicked() {
  switch (state) {
   case 19:
    state=1;
    break;
   case 20:
    state=0;
    break;
   case 100:
    state=1;
    break;
   default:
    state--;
    break;
  }
  analyzeState();
 }
 public synchronized void doneClicked() {
  stop();
  start();
 }
 public double[][] createCube() {
  double[][] ret=new double[8][3];
  ret[0][0]=-1;		ret[0][1]=-1;		ret[0][2]=-1;
  ret[1][0]=-1;		ret[1][1]=-1;		ret[1][2]=1;
  ret[2][0]=-1;		ret[2][1]=1;		ret[2][2]=-1;
  ret[3][0]=-1;		ret[3][1]=1;		ret[3][2]=1;
  ret[4][0]=1;		ret[4][1]=-1;		ret[4][2]=-1;
  ret[5][0]=1;		ret[5][1]=-1;		ret[5][2]=1;
  ret[6][0]=1;		ret[6][1]=1;		ret[6][2]=-1;
  ret[7][0]=1;		ret[7][1]=1;		ret[7][2]=1;
  return scaleBy(ret,0.5);
 }
 public static void logCube(String pre,double[][] cube) {
  for (int i=0;i<8;i++) {
   System.out.println(pre+": ("+cube[i][0]+","+cube[i][1]+","+cube[i][2]+")");
  }
 }
 public static double[][] cloneCube(double[][] cube) {
  double[][] ret=new double[8][3];
  for (int i=0;i<8;i++) for (int j=0;j<3;j++) ret[i][j]=cube[i][j];
  return ret;
 }
 public double[][] randomizeOrientation(double[][] _cube) {
  double fi;
  double[][] cube=cloneCube(_cube);
  fi=rand.nextDouble()*2.0*Math.PI;
  for (int i=0;i<8;i++) {
   double x=cube[i][0],y=cube[i][1];
   cube[i][0]=(x*Math.cos(fi))-(y*Math.sin(fi));
   cube[i][1]=(x*Math.sin(fi))+(y*Math.cos(fi));
  }
  fi=rand.nextDouble()*2.0*Math.PI;
  for (int i=0;i<8;i++) {
   double y=cube[i][1],z=cube[i][2];
   cube[i][2]=(y*Math.cos(fi))-(z*Math.sin(fi));
   cube[i][1]=(y*Math.sin(fi))+(z*Math.cos(fi));
  }
  fi=rand.nextDouble()*2.0*Math.PI;
  for (int i=0;i<8;i++) {
   double x=cube[i][0],z=cube[i][2];
   cube[i][0]=(x*Math.cos(fi))-(z*Math.sin(fi));
   cube[i][2]=(x*Math.sin(fi))+(z*Math.cos(fi));
  }
  return cube;
 }
 public double cmToUnit(double x) { return x*2/interOcularDist; }
 public double unitToCm(double x) { return x*interOcularDist/2; }
 public double[][] scaleBy(double[][] _cube,double scale) {
  // scale will most likely be cubeSize*2/interOcularDist
  double[][] cube=cloneCube(_cube);
  for (int i=0;i<8;i++) {
   for (int j=0;j<3;j++) cube[i][j]=cube[i][j]*scale;
  }
  return cube;
 }
 public double[][] translateBy(double[][] _cube,double x,double y,double z) {
  // user will specify x/y/z
  // x/y/z should be multiplied by 2, and divided by interOcularDist
  double[][] cube=cloneCube(_cube);
  for (int i=0;i<8;i++) {
   cube[i][0]+=x;
   cube[i][1]+=y;
   cube[i][2]+=z;
  }
  return cube;
 }
 public double[] createVisualAxis(double whichEye,double posX,double posY,
  double posZ) {
  double[] ret=new double[3];
  ret[0]=posX-whichEye;
  ret[1]=posY;
  ret[2]=posZ;
  System.out.println("Visual-Axis Vector: ("+ret[0]+","+ret[1]+","+ret[2]+")");
  return ret;
 }
 public double computeRetinalPlane(double[] visAx,double whichEye) {
  return ((0.0-whichEye)*visAx[0])-Math.sqrt(Math.pow(visAx[0],2.0)+
   Math.pow(visAx[1],2.0)+Math.pow(visAx[2],2.0));
 }
 public double[][] findIntersects(double[] visAx,double planConst,
  double whichEye,double[][] cube) {
  double[][] ret=new double[8][3];
  double
   negWhichEye=0.0-whichEye,	// 1.0 or -1.0
   axX=visAx[0],
   axY=visAx[1],
   axZ=visAx[2];
  for (int i=0;i<8;i++) {
   double
    ptX=cube[i][0],
    ptY=cube[i][1],
    ptZ=cube[i][2],
    retX,retY,retZ,theMagicNumber;
   theMagicNumber=
    ((negWhichEye*axX)-planConst)/
    ((ptX*axX)+(negWhichEye*axX)+(ptY*axY)+(ptZ*axZ));
   retX=whichEye+(theMagicNumber*(ptX+negWhichEye));
   retY=theMagicNumber*ptY;
   retZ=theMagicNumber*ptZ;
   ret[i][0]=retX;
   ret[i][1]=retY;
   ret[i][2]=retZ;
  }
  return ret;
 }
 public double[] computeCrossProduct(double[] a,double[] b) {
  double[] ret=new double[3];
  ret[0]=(a[1]*b[2])-(a[2]*b[1]);
  ret[1]=(a[2]*b[0])-(a[0]*b[2]);
  ret[2]=(a[0]*b[1])-(a[1]*b[0]);
  return ret;
 }
 public double computeMagnitude(double[] x) {
  return Math.sqrt((x[0]*x[0])+(x[1]*x[1])+(x[2]*x[2]));
 }
 public double[] computeUnitVector(double[] x) {
  double[] ret=new double[3];
  double mag=computeMagnitude(x);
  System.out.println("mag="+mag);
  ret[0]=x[0]/mag;
  ret[1]=x[1]/mag;
  ret[2]=x[2]/mag;
  return ret;
 }
 public double computeDotProduct(double[] a,double[] b) {
  double ret=0.0;
  for (int i=0;i<3;i++) ret+=a[i]*b[i];
  return ret;
 }
 public double[] computeRetinaRotationAxis(double[] visAx,
  double planConts) {
  return computeUnitVector(computeCrossProduct(visAx,zAxis));
 }
 public double getRotationAngle(double[] visAx) {
  double xCos=computeDotProduct(visAx,zAxis)/computeMagnitude(visAx);
  double xTan=Math.sqrt(1.0-(xCos*xCos))/xCos;
  return Math.abs(Math.atan(xTan));
 }
 public double[][] translateAndRotate(double[][] intersects,double[] visAx,
  double planConst,double whichEye,double rotAng,double[] rotAxis) {
  double[][] newInter=new double[8][3];
  for (int i=0;i<8;i++) {
   newInter[i][0]=intersects[i][0]-whichEye;
   newInter[i][1]=intersects[i][1];
   newInter[i][2]=intersects[i][2];
  }
  for (int i=0;i<8;i++) {
   double[] point=new double[3];
   point[0]=newInter[i][0];
   point[1]=newInter[i][1];
   point[2]=newInter[i][2];
   double[] cross2=computeCrossProduct(rotAxis,point);
   double[] cross3=computeCrossProduct(rotAxis,cross2);
   System.out.println("cross2=("+cross2[0]+","+cross2[1]+","+cross2[2]+")");
   System.out.println("cross3=("+cross3[0]+","+cross3[1]+","+cross3[2]+")");
   for (int j=0;j<3;j++)
    newInter[i][j]=point[j]+(cross2[j]*Math.sin(rotAng))+
    (2*cross3[j]*Math.pow(Math.sin(rotAng/2.0),2.0));
  }
  return newInter;
 }
 public double nextGaussian() {
  /*int irnd1=Math.abs(rand.nextInt()%100000);
  int irnd2=Math.abs(rand.nextInt()%100000);
  double rnd1=((double)irnd1)/100000;
  double rnd2=((double)irnd2)/100000;
  return Math.sqrt(-2.0*Math.log(rnd1))*Math.cos(2.0*Math.PI*rnd2);*/
  return rand.nextGaussian();
 }
 public double[][] addNoise(double[][] _intersects,double stdDevGaussian) {
  // stdDevGaussian should be 2% by default
  System.out.println("stdDevGaussian: "+stdDevGaussian);
  double[][] intersects=cloneCube(_intersects);
  for (int i=0;i<8;i++) {
   double ddx=Math.sqrt(
    (intersects[i][0]*intersects[i][0])+
    (intersects[i][1]*intersects[i][1]));
   System.out.println("ddx="+ddx);
   System.out.println("XY values: ("+intersects[i][0]+","+intersects[i][1]+")");
   intersects[i][0]+=(stdDevGaussian*ddx*nextGaussian());
   intersects[i][1]+=(stdDevGaussian*ddx*nextGaussian());
   System.out.println("XY values: ("+intersects[i][0]+","+intersects[i][1]+")");
  }
  return intersects;
 }
 public double[][] addUniversalError(double[][] _newInter,double[] axis,
  double ve) {
  // vergence: y-axis
  // elevation: x-axis
  // torsion: z-axis
  // ve is the error!
  // newInter must be the points that are already rotated and translated, and,
  // potentially, noisified
  double[][] newInter=cloneCube(_newInter);
  for (int i=0;i<8;i++) {
   double[] cross2=computeCrossProduct(axis,newInter[i]);
   double[] cross3=computeCrossProduct(axis,cross2);
   for (int j=0;j<3;j++)
    newInter[i][j]=newInter[i][j]+(cross2[j]*Math.sin(ve))+
    (2*cross3[j]*Math.pow(Math.sin(ve/2.0),2.0));
  }
  return newInter;
 }
 public double[][] undoTranslateAndRotate(double[][] intersects,double[] visAx,
  double planConst,double whichEye,double rotAng,double[] rotAxis) {
  double[][] newInter=new double[8][3];
  for (int i=0;i<8;i++) {
   double[] point=new double[3];
   point[0]=intersects[i][0];
   point[1]=intersects[i][1];
   point[2]=intersects[i][2];
   double[] cross2=computeCrossProduct(rotAxis,point);
   double[] cross3=computeCrossProduct(rotAxis,cross2);
   for (int j=0;j<3;j++)
    newInter[i][j]=point[j]-(cross2[j]*Math.sin(rotAng))+
    (2*cross3[j]*Math.pow(Math.sin(rotAng/2.0),2.0));
  }
  for (int i=0;i<8;i++) {
   newInter[i][0]=newInter[i][0]+whichEye;
   newInter[i][1]=newInter[i][1];
   newInter[i][2]=newInter[i][2];
  }
  return newInter;
 }
 double[][] reconstructCube(double[][] intersectsLeft,
  double[][] intersectsRight) {
  double[][] ret=new double[8][3];
  double[] left=new double[3],right=new double[3];
  double[] republican=new double[3],democrat=new double[3];
  for (int i=0;i<8;i++) {
   left[0]=intersectsLeft[i][0]-leftEye;
   left[1]=intersectsLeft[i][1];
   left[2]=intersectsLeft[i][2];
   right[0]=intersectsRight[i][0]-rightEye;
   right[1]=intersectsRight[i][1];
   right[2]=intersectsRight[i][2];
   double uglyNum=(2*left[0]*((right[1]*left[1])+(right[2]*left[2])))-
    (2*right[0]*((left[1]*left[1])+(left[2]*left[2])));
   double uglyDen=
    (((right[0]*right[0])+(right[1]*right[1])+(right[2]*right[2]))*
    ((left[0]*left[0])+(left[1]*left[1])+(left[2]*left[2])))-
    Math.pow(computeDotProduct(left,right),2.0);
   double uglyRatio=uglyNum/uglyDen;
   double prettyRatio=
    ((2*right[0])+(uglyRatio*computeDotProduct(right,right)))/
    computeDotProduct(right,left);
   republican[0]=leftEye+(prettyRatio*left[0]);
   republican[1]=prettyRatio*left[1];
   republican[2]=prettyRatio*left[2];
   democrat[0]=rightEye+(uglyRatio*right[0]);
   democrat[1]=uglyRatio*right[1];
   democrat[2]=uglyRatio*right[2];
   ret[i][0]=(republican[0]+democrat[0])/2;
   ret[i][1]=(republican[1]+democrat[1])/2;
   ret[i][2]=(republican[2]+democrat[2])/2;
  }
  return ret;
 }
}









